#include <iostream>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <CL/cl.h>

int main(int argc, char **argv) {

	int N = 2000;
	int status = 0;
   //Get the number of opencl platforms
   cl_uint numPlatforms;
   clGetPlatformIDs (0, NULL, &numPlatforms);
   //Get the opencl platform ids
   cl_platform_id* PlatformIDs;
   PlatformIDs = (cl_platform_id*)malloc(sizeof(cl_platform_id) * numPlatforms);
   clGetPlatformIDs(numPlatforms, PlatformIDs, NULL);

   //Get a GPU device
   //Note this is my own code so may be buggy
   cl_device_id device_id;
   int i;
   for (i = 0; i < numPlatforms; i++) {
      cl_int err = clGetDeviceIDs(PlatformIDs[i], CL_DEVICE_TYPE_GPU, 1, &device_id, NULL);
      if (err == CL_SUCCESS) {
         break;
      }
   }
   if (numPlatforms == i) {
      printf("Could not find GPU\n");
      exit(1);
   }
   
   //Create a context
   cl_context hContext;
   hContext = clCreateContext(0, 1, &device_id, NULL, NULL, NULL);
   if (!hContext) {
      printf("Error: no context\n");
      exit(1);
   }
   //Create a command queue for the device in the context
   cl_command_queue hCmdQueue;
   hCmdQueue = clCreateCommandQueue(hContext, device_id, CL_QUEUE_PROFILING_ENABLE, NULL);
   if (!hCmdQueue) {
      printf("Error: no command queue\n");
      exit(1);
   }
   //My code - read the opencl program into a string
   FILE *f = fopen("cltest.cl", "r");
   if (!f) {
      printf("Couldn't open opencl code\n");
      exit(1);
   }
   i = 0;
   char *program = (char*)malloc(sizeof(char) * 10000);
   memset(program, 0, sizeof(char) *10000);
   while(!feof(f)) {
      program[i] = fgetc(f);
      if (program[i] == EOF) {
      //if (program[i] == '@') {
         program[i] = 0;
         break;
      }
      i++;
   }
   //program[i] = NULL;
   fclose(f);
   //Create and compile the opencl program
   cl_program hProgram;
   hProgram = clCreateProgramWithSource(hContext, 1, (const char **)&program, 0, 0);
   if (!hProgram) {
      printf("Couldn't create program\n");
      exit(1);
   }
   int err = clBuildProgram(hProgram, 0, 0, 0, 0, 0);
   if (err != CL_SUCCESS) {
      printf("Error: couldn't build program\n");
      size_t len;
      char buffer[10000];
      clGetProgramBuildInfo(hProgram, device_id, CL_PROGRAM_BUILD_LOG, sizeof(char) * 10000, buffer, &len);
      printf("%s", buffer); 
      exit(1);
   }
   //Create the kernel instance
   cl_kernel hKernel;
   hKernel = clCreateKernel(hProgram, "fun", &err);
   if (!hKernel) {
      printf("Error: couldn't create kernel\n");
      exit(1);
   }
   cl_kernel hKernel2;
   hKernel2 = clCreateKernel(hProgram, "fun2", &err);
   if (!hKernel2) {
      printf("Error: couldn't create kernel\n");
      exit(1);
   }

   //Allocate and initialise the host array
   float *hostIn;
   float *hostOut;
   int size = sizeof(float) * N;

   srand(time(NULL));

   hostIn = (float *)malloc(size);
   hostOut = (float *)malloc(size);
   for (i = 0; i < N; i++) {
      hostOut[i] = 0.1;
   }

   //Allocate and initialise the gpu array
   cl_mem deviceIn = clCreateBuffer(hContext, CL_MEM_READ_WRITE, 
             N * sizeof(cl_float), 0, 0);
   cl_mem deviceOut = clCreateBuffer(hContext, CL_MEM_READ_WRITE,
         N * sizeof(cl_float), 0, 0);
   if (!deviceIn || !deviceOut) {
      printf("Error: couldn't allocate memory\n");
      exit(1);
   }
   clEnqueueWriteBuffer(hCmdQueue, deviceOut, CL_FALSE, 0, size, hostOut, 0, 0, 0);

   //Set the arguments to the kernel
   clSetKernelArg(hKernel, 0, sizeof(cl_mem), &deviceIn);
   clSetKernelArg(hKernel, 1, sizeof (int), &N);
   clSetKernelArg(hKernel2, 0, sizeof(cl_mem), &deviceIn);
   clSetKernelArg(hKernel2, 1, sizeof(cl_mem), &deviceOut);
   clSetKernelArg(hKernel2, 2, sizeof (int), &N);
clFinish(hCmdQueue);
	
int iterations = 1000;
	cl_ulong submit_start_time[4] = {0,0,0,0};
   for (i = 0; i < iterations; i++) {
	cl_event events[4];
      int j;
      for (j = 0; j < N; j++) {
         hostIn[j] = rand() / (float) RAND_MAX;
      }
      status = clEnqueueWriteBuffer(hCmdQueue, deviceIn, CL_TRUE, 0, size, hostIn, 0, 0, &events[0]);
	if(status != CL_SUCCESS)
	{
		std::cout << " Error in WriteBuffer" << status;
		exit(1);
	}
      clFlush(hCmdQueue);
      //Launch the kernel
      size_t global = 2048;
      size_t local = 256;
      status = clEnqueueNDRangeKernel(hCmdQueue, hKernel, 1, 0, &global, &local, 0, 0, &events[1]);
	if(status != CL_SUCCESS)
	{
		std::cout << " Error in kernel1)" << status;
		exit(1);
	}
      clFinish(hCmdQueue);
      status = clEnqueueNDRangeKernel(hCmdQueue, hKernel2, 1, 0, &global, &local, 0, 0, &events[2]);
	if(status != CL_SUCCESS)
	{
		std::cout << " Error in kernel2" << status;
		exit(1);
	}
      clFinish(hCmdQueue);

      //Copy results back to host
      status = clEnqueueReadBuffer(hCmdQueue, deviceOut, CL_TRUE, 0, N * sizeof(float),
            hostOut, 0, 0, &events[3]);
	if(status != CL_SUCCESS)
	{
		std::cout << " Error in Readbuffer " << status;
		exit(1);
	}
      status = clFinish(hCmdQueue);
     
	clFinish(hCmdQueue);
      for(int j=0; j<4; j++)
      {
	cl_ulong start_time, submit_time;
	status = clGetEventProfilingInfo(events[j], CL_PROFILING_COMMAND_START, sizeof(cl_ulong), &start_time, NULL);
	if(status != CL_SUCCESS)
	{
		std::cout << " Error in getting Profiling Info (start time)" << status;
		exit(1);
	}
	status = clGetEventProfilingInfo(events[j], CL_PROFILING_COMMAND_SUBMIT, sizeof(cl_ulong), &submit_time, NULL);
	if(status != CL_SUCCESS)
	{
		std::cout << " Error in getting Profiling Info (submit time)" << status;
		exit(1);
	}
	submit_start_time[j] += start_time - submit_time;
	}
   }

	clFinish(hCmdQueue);
	std::cout << "Average Launch Time for WriteBuffer Command(ms):" << (double)submit_start_time[0] / (iterations * 1e6)
		<< "\n Average Launch Time for kernel1 Command(ms):" << (double)submit_start_time[1] / (iterations * 1e6)
		<< "\n Average Launch Time for kernel2 Command(ms):" << (double)submit_start_time[2] / (iterations * 1e6)
		<< "\n Average Launch Time for ReadBuffer Command(ms):" << (double)submit_start_time[3] / (iterations * 1e6) << std::endl;
printf("Done");
   clReleaseKernel(hKernel);
   clReleaseProgram(hProgram);
   clReleaseMemObject(deviceIn);
   clReleaseMemObject(deviceOut);
   clReleaseCommandQueue(hCmdQueue);
   clReleaseContext(hContext);

   return 0;
}


